from scidb.core import Bucket, DataSet, Data
from .iteration import iter_data_set, iter_data
from typing import Set, Callable, Union


def search_data_set(
        bucket_or_data_set: Union[Bucket, DataSet],
        name_or_uuid: str,
        compare_func: Union[None, Callable] = None) -> Set[DataSet]:
    def search_ds(ds: DataSet, r: Set):
        if compare_func is None:
            if ds.name == name_or_uuid or ds.uuid == name_or_uuid:
                r.add(ds)
        else:
            if compare_func(ds, name_or_uuid):
                r.add(ds)
    results = set()
    iter_data_set(bucket_or_data_set, search_ds, r=results)
    return results


def search_data(
        bucket_or_data_set: Union[Bucket, DataSet],
        name: str,
        compare_func: Union[None, Callable] = None) -> Set[Data]:
    def search_d(d: Data, r: Set):
        if compare_func is None:
            if d.name == name:
                r.add(d)
        else:
            if compare_func(d, name):
                r.add(d)
    results = set()
    iter_data_set(bucket_or_data_set, search_d, r=results)
    return results



"""
Command list:

dataset create <name>
dataset list [all | deleted]
dataset cd [<name> | <uuid> | .. | parent ]
dataset rm [<name> | <uuid>]
dataset restore [<name> | <uuid> | all]
dataset clean [-f] [<name> | <uuid> | all]
dataset tree [<name> | <uuid>]
dataset search [<name> | <uuid>]
"""

from scidb.core import Bucket, DataSet
from typing import List, Set, Union
import scidb.client.global_env as global_env


usage = """\
 1 | > dataset create <name>
   |   To create a dataset with given name in current directory.
 2 | > dataset list [all | deleted]
   |   To list (all / deleted) data sets of current directory.
 3 | > dataset cd [<name> | <uuid> | .. | parent ]
   |   To navigate into / out of a directory.
 4 | > dataset rm [<name> | <uuid>]
   |   To delete a dataset with given name or uuid.
 5 | > dataset restore [<name> | <uuid> | all]
   |   To restore a deleted dataset.
 6 | > dataset clean [-f] [<name> | <uuid> | all]
   |   To delete trash of a dataset / all data sets.
 7 | > dataset tree [<name> | <uuid> | . ]
   |   To print a tree of a given dataset or current directory.
 8 | > dataset search [<name> | <uuid>]
   |   To search a dataset within current bucket.
"""

create_usage = """\
> dataset create <name>

To create a dataset with given name in current directory.
<name> | REQUIRED | name of the new dataset.
"""

list_usage = """\
> dataset list [all | deleted]

To list (all / deleted) data sets of current directory.
all     | OPTIONAL | list all data sets.
deleted | OPTIONAL | list deleted data sets.
"""

cd_usage = """\
dataset cd [<name> | <uuid> | .. | parent ]

To navigate into / out of a directory.
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset,
..     | OPTIONAL | parent directory,
parent | OPTIONAL | parent directory.
"""

rm_usage = """\
> dataset rm [<name> | <uuid>]

To delete a dataset with given name or uuid.
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset.
"""

restore_usage = """\
> dataset restore [<name> | <uuid> | all]

To restore a deleted dataset.
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset,
all    | OPTIONAL | restore all data sets.
"""

clean_usage = """\
> dataset clean [-f] [<name> | <uuid> | all]

To delete trash of a dataset / all data sets.
-f     | OPTIONAL | force clean (without confirmation),
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset,
all    | OPTIONAL | clean all data sets.
"""

tree_usage = """\
> dataset tree [<name> | <uuid> | . ]

To print a tree of a given dataset or current directory.
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset,
.      | OPTIONAL | current directory.
"""

search_usage = """\
> dataset search [<name> | <uuid>]

To search a dataset within current bucket.
<name> | OPTIONAL | name of the dataset,
<uuid> | OPTIONAL | uuid of the dataset.
"""


def handler(args: List[str]):
    if len(args) < 1:
        print(usage)
        return
    if global_env.SELECTED_BUCKET is None:
        print('No bucket selected.')
        return
    if not isinstance(global_env.SELECTED_BUCKET, Bucket):
        print('Internal error.')
        exit(-1)
        return
    if args[0] == 'create':
        if len(args) != 2:
            print(create_usage)
            return
        create_data_set(args[1])
    elif args[0] == 'list':
        if len(args) == 1:
            list_data_set()
        elif len(args) == 2 and args[1] in ['all', 'deleted']:
            list_data_set(args[1])
        else:
            print(list_usage)
            return
    elif args[0] == 'cd':
        if len(args) != 2:
            print(cd_usage)
            return
        cd_data_set(args[1])
    elif args[0] == 'rm':
        if len(args) != 2:
            print(rm_usage)
            return
        rm_data_set(args[1])
    elif args[0] == 'restore':
        if len(args) != 2:
            print(restore_usage)
            return
        if 'all' in args:
            restore_data_set()
        else:
            restore_data_set(args[1])
    elif args[0] == 'clean':
        if len(args) not in [2, 3]:
            print(clean_usage)
            return
        if '-f' in args:
            args.remove('-f')
            if args[1] == 'all':
                clean_data_set(None, confirm=True, feedback=True)
            else:
                clean_data_set(args[1], confirm=True, feedback=True)
        else:
            if args[1] == 'all':
                clean_data_set(None, confirm=True,
                               feedback=input('Would you like to clean all data sets? Y/[N]') == 'Y')
            else:
                clean_data_set(args[1], confirm=True,
                               feedback=input('Would you like to clean this dataset? Y/[N]') == 'Y')
    elif args[0] == 'tree':
        if len(args) != 2:
            print(tree_usage)
            return
        print_tree_of_data_set(args[1])
    elif args[0] == 'search':
        if len(args) != 2:
            print(search_usage)
            return
        search_data_set(args[1])
    else:
        print(usage)
        return


def get_parent() -> Union[Bucket, DataSet]:
    return global_env.SELECTED_BUCKET if global_env.CURRENT_DATASET is None else global_env.CURRENT_DATASET


def create_data_set(data_set_name: str):
    get_parent().add_data_set(data_set_name)


def print_data_sets(data_sets: Set[DataSet]):
    print('  No. |  dataset name   |                 uuid                ')
    print('--------------------------------------------------------------')
    for i, data_set in enumerate(data_sets):
        print(f'{i: 5} | {data_set.name: <15} | {data_set.uuid}')
    if len(data_sets) == 0:
        print('                         No records.                          ')
    print('--------------------------------------------------------------')


def list_data_set(data_set_filter: Union[None, str] = None):
    if data_set_filter is None:
        print_data_sets(get_parent().data_sets)
    elif data_set_filter == 'all':
        print_data_sets(get_parent().all_data_sets)
    elif data_set_filter == 'deleted':
        print_data_sets(get_parent().trash)
    else:
        return


def cd_data_set(target: str):
    if target in ['..', 'parent']:
        if isinstance(get_parent(), Bucket):
            print('You cannot switch above buckets.')
        else:
            parent = get_parent().parent
            if isinstance(parent, DataSet):
                global_env.CURRENT_DATASET = parent
            else:
                global_env.CURRENT_DATASET = None
    else:
        data_set = get_parent().get_data_set(target)
        if data_set is not None:
            global_env.CURRENT_DATASET = data_set
        else:
            print('No such dataset.')


def rm_data_set(name_or_uuid: str):
    data_set = get_parent().get_data_set(name_or_uuid)
    if data_set is not None:
        data_set.delete()


def restore_data_set(name_or_uuid: Union[str, None] = None):
    if name_or_uuid is None:
        for deleted in get_parent().trash:
            deleted.restore()
    else:
        target = get_parent().get_data_set(name_or_uuid, include_deleted=True)
        if target is not None:
            target.restore()
        else:
            print('No such dataset.')


def clean_data_set(name_or_uuid: Union[str, None], confirm: bool = True, feedback: bool = False):
    if confirm and not feedback:
        print('User cancelled.')
        return
    if name_or_uuid is None:
        get_parent().clear_trash(confirm, feedback)
    else:
        data_set = get_parent().get_data_set(name_or_uuid)
        if data_set is None:
            print('No such dataset.')
        else:
            data_set.clear_trash(confirm, feedback)


def print_tree(data_set: DataSet, depth: int = 0):
    if depth > 0:
        prefix = ' |  ' * (depth - 1) + ' |--'
    else:
        prefix = ''
    print(f'{prefix} # {data_set.name} ({data_set.uuid})')
    for child in data_set.data_sets:
        print_tree(child, depth=depth + 1)


def print_tree_of_data_set(name_or_uuid: str):
    if name_or_uuid == '.':
        data_set = get_parent()
    else:
        data_set = get_parent().get_data_set(name_or_uuid)
    if data_set is None:
        print('No such dataset.')
    else:
        print_tree(data_set)


def print_search_results(target: Union[None, DataSet], path: List):
    if target is None:
        return
    print(f'@ Found result: {target.name} ({target.uuid})')
    print('  path:')
    path_str = [f'{item.name} ({item.uuid})' for item in path]
    path_str = ' --> \n'.join(path_str)
    print(path_str)


def search_data_set(name_or_uuid: str,
                    parent: Union[None, Bucket, DataSet] = None,
                    path: Union[None, List] = None):
    if path is None:
        path = []
    if parent is None:
        parent = global_env.SELECTED_BUCKET
    path.append(parent)
    target = parent.get_data_set(name_or_uuid)
    if target is not None:
        path.append(target)
        print_search_results(target, path)
    else:
        for child in parent.data_sets:
            search_data_set(name_or_uuid, child, path)



import shutil
import hashlib
from pathlib import Path
from typing import TextIO, BinaryIO, IO, Union
from .low import ObservableDict


class Data:
    def __init__(self, data_name: str, parent):
        self.__data_name__ = data_name
        self.__parent__ = parent
        self.init_metadata()
        self.init_properties()

    def init_metadata(self):
        if self.__data_name__ not in self.__parent__.metadata:
            self.__parent__.metadata[self.__data_name__] = dict()

    def init_properties(self):
        if self.__data_name__ not in self.__parent__.properties:
            self.__parent__.properties[self.__data_name__] = dict()

    def set_metadata(self, metadata: Union[None, dict], merge: bool = True):
        if metadata is None:
            return
        if merge:
            metadata = {**self.metadata, **metadata}
        self.__parent__.metadata[self.__data_name__] = metadata

    def set_properties(self, properties: Union[None, dict], merge: bool = True):
        if properties is None:
            return
        if merge:
            properties = {**self.properties, **properties}
        self.__parent__.properties[self.__data_name__] = properties

    @property
    def parent(self):
        return self.__parent__

    @property
    def path(self) -> Path:
        return self.__parent__.path / self.__data_name__

    @property
    def name(self) -> str:
        return self.__data_name__

    @property
    def metadata(self) -> ObservableDict:
        return self.__parent__.metadata[self.__data_name__]

    @property
    def properties(self) -> ObservableDict:
        return self.__parent__.properties[self.__data_name__]

    def rename(self, new_name: str):
        shutil.move(str(self.path), str(self.__parent__.path / new_name))
        self.__data_name__ = new_name

    def reader(self, binary: bool = False, **kwargs) -> [IO, BinaryIO, TextIO, None]:
        mode = 'r'
        mode += 'b' if binary else ''
        return open(str(self.path), mode=mode, **kwargs)

    def creator(self,
                binary: bool = False,
                confirm: bool = False,
                feedback: bool = False,
                **kwargs) -> [IO, BinaryIO, TextIO, None]:
        if confirm and not feedback:
            return None
        mode = 'x'
        mode += 'b' if binary else ''
        return open(str(self.path), mode=mode, **kwargs)

    def writer(self,
               binary: bool = False,
               append: bool = True,
               allow_overwrite: bool = False,
               confirm: bool = True,
               feedback: bool = False,
               **kwargs) -> [IO, BinaryIO, TextIO, None]:
        if not allow_overwrite and not append:
            raise PermissionError('Trying to overwrite existed data.')
        if confirm and not feedback:
            return
        mode = 'a' if append else 'w'
        mode += 'b' if binary else ''
        return open(str(self.path), mode=mode, **kwargs)

    def __repr__(self):
        return f"Data('{self.__data_name__}')"

    def import_file(self, src_path: [str, Path], allow_overwrite=False, confirm=True, feedback=False):
        if self.path.exists() and not allow_overwrite:
            return
        if confirm and not feedback:
            return
        shutil.copyfile(str(src_path), str(self.path))

    def export_file(self, dst_path: [str, Path], allow_overwrite=False):
        if Path(dst_path).exists() and not allow_overwrite:
            return
        shutil.copyfile(str(self.path), str(dst_path))

    def __calc_hash__(self, h, buffer_size: int = 131072):
        if not self.path.exists():
            return None
        with open(str(self.path), 'rb') as file_reader:
            while True:
                data = file_reader.read(buffer_size)
                if not data:
                    break
                h.update(data)
        return h.hexdigest()

    def md5(self, buffer_size: int = 131072, require_update: bool = True) -> [str, None]:
        if require_update or 'md5' not in self.metadata:
            result = self.__calc_hash__(hashlib.md5(), buffer_size)
            self.metadata['md5'] = result
            return result
        else:
            return self.metadata['md5']

    def sha1(self, buffer_size: int = 131072, require_update: bool = True) -> [str, None]:
        if require_update or 'sha1' not in self.metadata:
            result = self.__calc_hash__(hashlib.sha1(), buffer_size)
            self.metadata['sha1'] = result
            return result
        else:
            return self.metadata['sha1']

    def sha256(self, buffer_size: int = 131072, require_update: bool = True) -> [str, None]:
        if require_update or 'sha256' not in self.metadata:
            result = self.__calc_hash__(hashlib.sha256(), buffer_size)
            self.metadata['sha256'] = result
            return result
        else:
            return self.metadata['sha256']



from scidb.core import Database, Bucket, DataSet
from typing import Union

CONNECTED_DATABASE: Union[None, Database] = None
SELECTED_BUCKET: Union[None, Bucket] = None
CURRENT_DATASET: Union[None, DataSet] = None



from .database import handler as db_handler
from .bucket import handler as bucket_handler
from .data_set import handler as data_set_handler
from .data import handler as data_handler
from .metadata import handler as metadata_handler



# Migrate
# database --> database
# bucket --> database
# bucket --> bucket
# bucket --> dataset
# dataset --> bucket
# dataset --> dataset
# data --> dataset
# data --> data

from scidb.core import Database, Bucket, DataSet, Data
from typing import Union


ALLOWED_MIGRATION = [
    (Database, Database),
    (Bucket, Database),
    (Bucket, Bucket),
    (Bucket, DataSet),
    (DataSet, Bucket),
    (DataSet, DataSet),
    (Data, DataSet),
    (Data, Data)
]


def migrate(source: Union[Database, Bucket, DataSet, Data],
            destination: Union[Database, Bucket, DataSet, Data],
            delete_source: bool = False,
            allow_overwrite: bool = False,
            confirm: bool = True,
            feedback: bool = False,
            verbose: bool = True):
    if not is_allowed(source, destination):
        raise AssertionError(f'Migration from {type(source)} to {type(destination)} is not allowed.')
    if confirm and not feedback:
        return
    if isinstance(source, Database):
        migrate_db(source, destination, delete_source, allow_overwrite, confirm, feedback, verbose)
    elif isinstance(source, Bucket):
        migrate_bucket(source, destination, delete_source, allow_overwrite, confirm, feedback, verbose)
    elif isinstance(source, DataSet):
        migrate_data_set(source, destination, delete_source, allow_overwrite, confirm, feedback, verbose)
    elif isinstance(source, Data):
        migrate_data(source, destination, delete_source, allow_overwrite, confirm, feedback)
    else:
        raise TypeError


def is_allowed(source: Union[Database, Bucket, DataSet, Data],
               destination: Union[Database, Bucket, DataSet, Data]):
    src_dst_types = type(source), type(destination)
    return src_dst_types in ALLOWED_MIGRATION


def migrate_db(source: Database,
               destination: Database,
               delete_source: bool = False,
               allow_overwrite: bool = False,
               confirm: bool = True,
               feedback: bool = False,
               verbose: bool = True):
    if confirm and not feedback:
        return
    for src_bucket in source.all_buckets:
        if not allow_overwrite and (
            destination.get_bucket(src_bucket.name, include_deleted=True)
            or destination.get_bucket(src_bucket.uuid, include_deleted=True)
        ):
            continue
        if verbose:
            print('Migrating:', src_bucket.node_type, src_bucket.name)
        if delete_source:
            src_bucket.move_to(destination)
        else:
            src_bucket.copy_to(destination)


def migrate_bucket(source: Bucket,
                   destination: Union[Database, Bucket, DataSet],
                   delete_source: bool = False,
                   allow_overwrite: bool = False,
                   confirm: bool = True,
                   feedback: bool = False,
                   verbose: bool = True):
    if confirm and not feedback:
        return
    if isinstance(destination, Database):
        if not allow_overwrite and (
            destination.get_bucket(source.name, include_deleted=True)
            or destination.get_bucket(source.uuid, include_deleted=True)
        ):
            return
        if verbose:
            print('Migrating:', source.node_type, source.name)
        if delete_source:
            source.move_to(destination)
        else:
            source.copy_to(destination)
    elif isinstance(destination, Bucket) or isinstance(destination, DataSet):
        if allow_overwrite:
            destination.set_metadata(source.metadata)
            destination.set_properties(source.properties)
        for src_bucket in source.all_data_sets:
            if not allow_overwrite and (
                destination.get_data_set(src_bucket.name, include_deleted=True)
                or destination.get_data_set(src_bucket.uuid, include_deleted=True)
            ):
                continue
            if verbose:
                print('Migrating:', src_bucket.node_type, src_bucket.name)
            if delete_source:
                src_bucket.move_to(destination)
            else:
                src_bucket.copy_to(destination)
        if delete_source:
            source.delete()
            source.parent.clear_trash(confirm, feedback)
    else:
        raise TypeError


def migrate_data_set(source: DataSet,
                     destination: Union[Bucket, DataSet],
                     delete_source: bool = False,
                     allow_overwrite: bool = False,
                     confirm: bool = True,
                     feedback: bool = False,
                     verbose: bool = True):
    if confirm and not feedback:
        return
    if not allow_overwrite and (
        destination.get_data_set(source.name, include_deleted=True)
        or destination.get_data_set(source.uuid, include_deleted=True)
    ):
        return
    if verbose:
        print('Migrating:', source.node_type, source.name)
    if delete_source:
        source.move_to(destination)
    else:
        source.copy_to(destination)


def migrate_data(source: Data,
                 destination: Union[DataSet, Data],
                 delete_source: bool = False,
                 allow_overwrite: bool = False,
                 confirm: bool = True,
                 feedback: bool = False,
                 verbose: bool = True):
    if confirm and not feedback:
        return
    if isinstance(destination, DataSet):
        if not allow_overwrite and destination.get_data(source.name):
            return
        if verbose:
            print('Migrating:', 'Data', source.name)
        data = destination.add_data(source.name)
        data.import_file(source.path, allow_overwrite, confirm, feedback)
        data.set_properties(source.properties)
        data.set_metadata(source.metadata)
        if delete_source:
            source.parent.delete_data(source.name, confirm, feedback)
    elif isinstance(destination, Data):
        if not allow_overwrite:
            return
        if verbose:
            print('Migrating:', 'Data', source.name)
        destination.import_file(source.path, allow_overwrite, confirm, feedback)
        destination.set_properties(source.properties)
        destination.set_metadata(source.metadata)
        if delete_source:
            source.parent.delete_data(source.name, confirm, feedback)
    else:
        raise TypeError



"""
Command list:

property edit [<name> | <uuid>] <key_path> <value>
property delete [<name> | <uuid>] <key_path>
property show [<name> | <uuid>]
metadata edit [<name> | <uuid>] <key_path> <value>
metadata delete [<name> | <uuid>] <key_path>
metadata show [<name> | <uuid>]
"""

from scidb.core import Database, Bucket, DataSet, Data
from typing import List, Any, Union
import scidb.client.global_env as global_env
from pprint import pprint


usage = """\
 1 | > property edit <node_type> [<name> | <uuid>] <key_path> <value>
   |   To create or edit an entry of property.
 2 | > property delete <node_type> [<name> | <uuid>] <key_path>
   |   To delete an entry of property.
 3 | > property show <node_type> [<name> | <uuid>]
   |   To show properties of an object.
 4 | > metadata edit <node_type> [<name> | <uuid>] <key_path> <value>
   |   To create or edit an entry of metadata.
 5 | > metadata delete <node_type> [<name> | <uuid>] <key_path>
   |   To delete an entry of metadata.
 6 | > metadata show <node_type> [<name> | <uuid>]
   |   To show metadata of an object.
"""

property_edit_usage = """\
> property edit <node_type> [<name> | <uuid>] <key_path> <value>

To create or edit an entry of property.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset,
<key_path>  | REQUIRED | path to a key,
<value>     | REQUIRED | property value.
"""

property_delete_usage = """\
> property delete <node_type> [<name> | <uuid>] <key_path>

To delete an entry of property.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset,
<key_path>  | REQUIRED | path to a key.
"""

property_show_usage = """\
> property show <node_type> [<name> | <uuid>]

To show properties of an object.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset.
"""

metadata_edit_usage = """\
> metadata edit <node_type> [<name> | <uuid>] <key_path> <value>

To create or edit an entry of metadata.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset,
<key_path>  | REQUIRED | path to a key,
<value>     | REQUIRED | property value.
"""

metadata_delete_usage = """\
> metadata delete <node_type> [<name> | <uuid>] <key_path>

To delete an entry of metadata.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset,
<key_path>  | REQUIRED | path to a key.
"""

metadata_show_usage = """\
> metadata show <node_type> [<name> | <uuid>]

To show metadata of an object.
<node_type> | REQUIRED | to specify a bucket / dataset or data,
<name>      | OPTIONAL | name of the dataset or data,
<uuid>      | OPTIONAL | uuid of the dataset.
"""


def handler(args: List[str]):
    if len(args) < 3:
        print(usage)
        return
    if global_env.SELECTED_BUCKET is None:
        print('No bucket selected.')
        return
    if not isinstance(global_env.SELECTED_BUCKET, Bucket):
        print('Internal error.')
        exit(-1)
        return
    if args[0] not in ['p', 'property', 'm', 'metadata']:
        print(usage)
        return
    if args[1] == 'edit':
        if len(args) < 6:
            print(metadata_edit_usage if args[0] == 'metadata' else property_edit_usage)
            return
        edit_entry(args[0], args[2], args[3], args[4], args[5])
    elif args[1] == 'delete':
        if len(args) < 5:
            print(metadata_delete_usage if args[0] == 'metadata' else property_delete_usage)
            return
        delete_entry(args[0], args[2], args[3], args[4])
    elif args[1] == 'show':
        if len(args) < 4:
            print(metadata_show_usage if args[0] == 'metadata' else property_show_usage)
            return
        show_entry(args[0], args[2], args[3])
    else:
        print(usage)
        return


def get_item(node_type: str, name_or_uuid: str) -> Union[Bucket, DataSet, Data, None]:
    if node_type in ['bk', 'bucket'] and isinstance(global_env.CONNECTED_DATABASE, Database):
        return global_env.CONNECTED_DATABASE.get_bucket(name_or_uuid, include_deleted=True)
    elif node_type in ['ds', 'dataset'] and (
            isinstance(global_env.SELECTED_BUCKET, Bucket) or
            isinstance(global_env.CURRENT_DATASET, DataSet)):
        parent = global_env.SELECTED_BUCKET if global_env.CURRENT_DATASET is None else global_env.CURRENT_DATASET
        return parent.get_data_set(name_or_uuid, include_deleted=True)
    elif node_type in ['d', 'data'] and isinstance(global_env.CURRENT_DATASET, DataSet):
        return global_env.CURRENT_DATASET.get_data(name_or_uuid)
    else:
        return None


def edit_entry(entry_type: str, node_type: str, name_or_uuid: str, key_path: str, value: Any):
    target = get_item(node_type, name_or_uuid)
    if target is None:
        print('No such item')
        return
    if entry_type in ['m', 'metadata']:
        target.metadata.set_by_path(key_path.split('.'), value)
    elif entry_type in ['p', 'property']:
        target.properties.set_by_path(key_path.split('.'), value)
    else:
        print('No such entry.')
        return


def delete_entry(entry_type: str, node_type: str, name_or_uuid: str, key_path: str):
    target = get_item(node_type, name_or_uuid)
    if target is None:
        print('No such item')
        return
    if entry_type in ['m', 'metadata']:
        target.metadata.delete_by_path(key_path.split('.'))
    elif entry_type in ['p', 'property']:
        target.properties.delete_by_path(key_path.split('.'))
    else:
        print('No such entry.')
        return


def show_entry(entry_type: str, node_type: str, name_or_uuid: str):
    target = get_item(node_type, name_or_uuid)
    if target is None:
        print('No such item')
        return
    if entry_type in ['m', 'metadata']:
        pprint(target.metadata)
    elif entry_type in ['p', 'property']:
        pprint(target.properties)
    else:
        print('No such entry.')
        return



from setuptools import setup

setup(
   name='oxdc-scidb',
   version='0.1b10',
   description='A simple scientific database.',
   author='oxdc',
   author_email='projaias@outlook.com',
   url='https://github.com/oxdc/sci.db',
   packages=[
      'scidb',
      'scidb.core', 'scidb.core.low',
      'scidb.client', 'scidb.client.modules',
      'scidb.utils',
      'scidb.plugins',
      'scidb.plugins.backup', 'scidb.plugins.backup.base', 'scidb.plugins.backup.implementations',
      'scidb.plugins.sandbox'
   ],
   install_requires=['PyYAML', 'minio', 'urllib3']
)



"""
Command list:

db create <name> <path>
db connect <name> <path>
db close
db rm [-f] <name> <path>
"""

from scidb.core import Database
from typing import List, Union
from pathlib import Path
import scidb.client.global_env as global_env
import shutil


usage = """\
 1 | > db create <name> <path>
   |   To create a new database.
 2 | > db connect <name> <path>
   |   To connect an existing database.
 3 | > db close
   |   To close current connection.
 4 | > db rm [-f] <name> <path>
   |   To delete an existing database.
"""

create_usage = """\
> db create <name> <path>

To create a new database.
<name> | REQUIRED | name of the new database,
<path> | REQUIRED | absolute path to the root directory.
"""

connect_usage = """\
> db connect <name> <path>

To connect an existing database.
<name> | REQUIRED | name of the database,
<path> | REQUIRED | absolute path to the root directory.
"""

close_usage = """\
> db close

To close current connection.
"""

rm_usage = """\
db rm [-f] <name> <path>

To delete an existing database.
-f     | OPTIONAL | force delete (without confirmation), 
<name> | REQUIRED | name of the database,
<path> | REQUIRED | absolute path to the root directory.
"""


def handler(args: List[str]):
    if len(args) < 1:
        print(usage)
        return
    if args[0] == 'create':
        if len(args) != 3:
            print(create_usage)
            return
        create_db(args[1], args[2])
    elif args[0] == 'connect':
        if len(args) != 3:
            print(connect_usage)
            return
        connect_db(args[1], args[2])
    elif args[0] == 'close':
        if len(args) != 1:
            print(close_usage)
            return
        close_db()
    elif args[0] == 'rm':
        if len(args) not in [3, 4]:
            print(rm_usage)
            return
        if '-f' in args:
            args.remove('-f')
            rm_db(args[1], args[2], confirm=False)
        else:
            rm_db(args[1], args[2], confirm=True,
                  feedback=input('Do you really want to delete this database? Y/[N]') == 'Y')
    else:
        print(usage)
        return


def create_db(name: str, path: Union[str, Path]):
    try:
        global_env.CONNECTED_DATABASE = Database(name, path)
    except FileExistsError:
        print('The directory is not empty. Please specify an empty one.')


def connect_db(name: str, path: Union[str, Path]):
    global_env.CONNECTED_DATABASE = Database(name, path)


def close_db():
    global_env.CONNECTED_DATABASE = None


def rm_db(name: str, path: Union[str, Path], confirm: bool = True, feedback: bool = False):
    if confirm and not feedback:
        print('User cancelled.')
        return
    shutil.rmtree(path)
    print(f'{name} deleted.')
